""" Main entry point of the Kheiron library """
from kheiron.opts import TrainingOptions

from kheiron.trainer import Trainer

__version__ = '1.0.4'
__author__ = 'TrongDat-Ngo'
__email__ = 'datnnt97@gmail.com'


__all__ = ['Trainer', 'TrainingOptions']

