"""Define constant values used across Kheiron."""


class SavedFiles(object):
    WEIGHT_MODEL_FILE = 'pytorch_model.bin'
    TRAINING_ARGS_FILE = 'training_args.bin'
    TRAINING_STATS_FILE = 'training_stats.json'
    OPTIMIZER_FILE = 'optimizer.bin'
    SCHEDULER_FILE = 'scheduler.bin'
